"""
Types specific to import
"""
from typing import Dict, NewType, Optional

from typing_extensions import Literal, Required, TypedDict

from kili.services.types import LabelType

Classes = NewType("Classes", Dict[int, str])
LabelFormat = Literal["yolo_v4", "yolo_v5", "yolo_v7", "raw"]


class LabelToImport(TypedDict, total=False):
    """
    Typed dictionary that specifies the keys and types for the labels to import.
    """

    author_id: str
    label_asset_external_id: Optional[str]
    label_asset_id: Optional[str]
    label_type: Optional[LabelType]
    seconds_to_label: Optional[int]
    path: Required[str]



"""Service for exporting kili objects """

from typing import List, Optional

from typing_extensions import get_args

from kili.exceptions import NotFound
from kili.services.export.format.base import (
    ContentRepositoryParams,
    ExportParams,
    LoggerParams,
)
from kili.services.export.format.kili import KiliExporterSelector
from kili.services.export.format.yolo import YoloExporterSelector
from kili.services.export.types import ExportType, LabelFormat, SplitOption
from kili.services.types import LogLevel, ProjectId


def export_labels(  # pylint: disable=too-many-arguments
    kili,
    asset_ids: Optional[List[str]],
    project_id: ProjectId,
    export_type: ExportType,
    label_format: LabelFormat,
    split_option: SplitOption,
    single_file: bool,
    output_file: str,
    disable_tqdm: bool,
    log_level: LogLevel,
) -> None:
    """
    Export the selected assets into the required format, and save it into a file archive.
    """
    if kili.count_projects(project_id=project_id) == 0:
        raise NotFound(f"project ID: {project_id}")

    export_params = ExportParams(
        assets_ids=asset_ids,
        project_id=project_id,
        export_type=export_type,
        label_format=label_format,
        split_option=split_option,
        single_file=single_file,
        output_file=output_file,
    )

    logger_params = LoggerParams(
        disable_tqdm=disable_tqdm,
        level=log_level,
    )

    content_repository_params = ContentRepositoryParams(
        router_endpoint=kili.auth.api_endpoint,
        router_headers={
            "Authorization": f"X-API-Key: {kili.auth.api_key}",
        },
    )

    if label_format in get_args(LabelFormat):
        if label_format == "raw":
            exporter_selector = KiliExporterSelector()
        else:
            exporter_selector = YoloExporterSelector()
        exporter = exporter_selector.select_exporter(
            kili, logger_params, export_params, content_repository_params
        )
        exporter.export_project(kili, export_params, logger_params)
    else:
        raise ValueError(f'Label format "{label_format}" is not implemented or does not exist.')



# pylint: skip-file
"""Legacy code for importing assets, to be removed soon"""

import csv
import mimetypes
import os
from json import dumps
from typing import Any, List, Optional, Tuple, Union
from uuid import uuid4

from kili.authentication import KiliAuth
from kili.graphql.operations.asset.mutations import (
    GQL_APPEND_MANY_FRAMES_TO_DATASET,
    GQL_APPEND_MANY_TO_DATASET,
)
from kili.helpers import (
    check_file_mime_type,
    encode_base64,
    format_result,
    get_data_type,
    is_url,
)
from kili.orm import Asset
from kili.utils import bucket, pagination

from .base import LoggerParams, ProcessingParams, ProjectParams
from .types import AssetLike


class LegacyDataImporter:
    """Legacy Asset importer
    First switch form new list of dict paradigm to the old set of list one, then import assets"""

    def __init__(
        self,
        auth: KiliAuth,
        project_params: ProjectParams,
        processing_params: ProcessingParams,
        logger_params: LoggerParams,
    ):
        self.auth = auth
        self.project_id = project_params.project_id
        self.input_type = project_params.input_type

    def import_assets(self, assets: List[AssetLike]):
        def ternary(field: str, default=None) -> Optional[List[Any]]:
            if assets[0].get(field, None):
                return [asset.get(field, default) for asset in assets]
            else:
                return None

        content_array = ternary("content")
        json_content_array = ternary("json_content")
        external_id_array = ternary("external_id", uuid4().hex)
        status_array = ternary("status", "TODO")
        is_honeypot_array = ternary("is_honeypot", False)
        json_metadata_array = ternary("json_metadata", {})

        (properties_to_batch, upload_type, request,) = process_append_many_to_dataset_parameters(
            self.auth,
            self.input_type,
            content_array,
            external_id_array,
            is_honeypot_array,
            status_array,
            json_content_array,
            json_metadata_array,
        )

        def generate_variables(batch):
            if request == GQL_APPEND_MANY_FRAMES_TO_DATASET:
                payload_data = {
                    "contentArray": batch["content_array"],
                    "externalIDArray": batch["external_id_array"],
                    "jsonMetadataArray": batch["json_metadata_array"],
                    "uploadType": upload_type,
                }
            else:
                payload_data = {
                    "contentArray": batch["content_array"],
                    "externalIDArray": batch["external_id_array"],
                    "isHoneypotArray": batch["is_honeypot_array"],
                    "statusArray": batch["status_array"],
                    "jsonContentArray": batch["json_content_array"],
                    "jsonMetadataArray": batch["json_metadata_array"],
                }
            return {"data": payload_data, "where": {"id": self.project_id}}

        results = pagination._mutate_from_paginated_call(
            self, properties_to_batch, generate_variables, request
        )
        return format_result("data", results[0], Asset)


def encode_object_if_not_url(content, input_type):
    """
    Return the object if it is a url, else it should be a path to a file.
    In that case, the file is returned as a base64 string
    """
    if is_url(content):
        return content
    if check_file_mime_type(content, input_type):
        return encode_base64(content)
    return None


def process_frame_json_content(json_content):
    """
    Function to process individual json_content of VIDEO projects
    """
    if is_url(json_content):
        return json_content
    json_content_index = range(len(json_content))
    json_content_urls = [encode_object_if_not_url(content, "IMAGE") for content in json_content]
    return dict(zip(json_content_index, json_content_urls))


def get_file_mimetype(
    content_array: Union[List[str], None], json_content_array: Union[List[str], None]
) -> Union[str, None]:
    """
    Returns the mimetype of the first file of the content array
    """
    if json_content_array is not None:
        return None
    if content_array is None:
        return None
    if len(content_array) > 0:
        first_asset = content_array[0]
        if is_url(first_asset):
            return None
        if not os.path.exists(first_asset):
            return None
        return mimetypes.guess_type(first_asset)[0]
    return None


def process_json_content(
    input_type: str,
    content_array: List[str],
    json_content_array: Union[List[str], None],
):
    """
    Process the array of json_contents and upload json if not already hosted
    """
    if json_content_array is None:
        return [""] * len(content_array)
    processed_json_content_array = []
    for json_content in json_content_array:
        if is_url(json_content):
            processed_json_content_array.append(json_content)
        else:
            if input_type in ("FRAME", "VIDEO"):
                json_content = process_frame_json_content(json_content)
            processed_json_content_array.append(dumps(json_content))
    return processed_json_content_array


def upload_content(signed_url: str, content: str, input_type: str):
    """
    Upload the content to a bucket if it is either a local file or raw text given for a TEXT project
    Args:
        content: the content to process, can be a local file or raw text
        signed_url: a signed_url to possibily use to upload the content
        input_type: input type of the project
    """
    if os.path.exists(content) and check_file_mime_type(content, input_type):
        with open(content, "rb") as file:
            data = file.read()
        content_type = get_data_type(content)
        uploaded_content_url = bucket.upload_data_via_rest(signed_url, data, content_type)
        return uploaded_content_url
    elif not os.path.exists(content) and input_type == "TEXT":
        data = content
        content_type = "text/plain"
        uploaded_content_url = bucket.upload_data_via_rest(signed_url, data, content_type)
        return uploaded_content_url
    else:
        raise ValueError(f"File: {content} not found")


# pylint: disable=too-many-arguments
def process_and_store_content(
    input_type: str,
    content_array: List[str],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    auth: KiliAuth,
):
    """
    Process the array of contents and upload content if not already hosted
    """
    if input_type == "TIME_SERIES":
        return list(map(process_time_series, content_array))
    if json_content_array is not None:
        return content_array
    has_local_files = any(not is_url(content) for content in content_array)
    url_content_array = []
    if has_local_files:
        signed_urls = bucket.request_signed_urls(auth, len(content_array))
    for i, content in enumerate(content_array):
        url_content = (is_url(content) and content) or upload_content(
            signed_urls[i], content, input_type  # type:ignore X
        )
        url_content_array.append(url_content)
    return url_content_array


def process_time_series(content: str) -> Union[str, None]:
    """
    Process the content for TIME_SERIES projects: if it is a file, read the content
    and also check if the content corresponds to the expected format, else return None
    """
    delimiter = ","
    if os.path.isfile(content):
        if check_file_mime_type(content, "TIME_SERIES"):
            with open(content, "r", encoding="utf8") as csvfile:
                reader = csv.reader(csvfile, delimiter=delimiter)
                return process_csv_content(reader, file_name=content, delimiter=delimiter)
        return None

    reader = csv.reader(content.split("\n"), delimiter=",")
    return process_csv_content(reader, delimiter=delimiter)


def is_float(number: str) -> bool:
    """
    Check if a string can be converted to float
    """
    try:
        float(number)
        return True
    except ValueError:
        return False


def process_csv_content(reader, file_name=None, delimiter=",") -> Optional[str]:
    """
    Process the content of csv for time_series and check if it corresponds to the expected format
    """
    first_row = True
    processed_lines = []
    for row in reader:
        if not (len(row) == 2 and (first_row or (not first_row and is_float(row[0])))):
            print(
                f"""The content {file_name if file_name else row} does not correspond to the \
correct format: it should have only 2 columns, the first one being the timestamp \
(an integer or a float) and the second one a numeric value (an integer or a float, \
otherwise it will be considered as missing value). The first row should have the names \
of the 2 columns. The delimiter used should be ','."""
            )
            return None
        value = row[1] if (is_float(row[1]) or first_row) else ""
        processed_lines.append(delimiter.join([row[0], value]))
        first_row = False
    return "\n".join(processed_lines)


def add_video_parameters(json_metadata, should_use_native_video):
    """
    Add necessary video parameters to the metadata of the video
    """
    processing_parameters = json_metadata.get("processingParameters", {})
    video_parameters = [
        ("shouldKeepNativeFrameRate", should_use_native_video),
        ("framesPlayedPerSecond", 30),
        ("shouldUseNativeVideo", should_use_native_video),
    ]
    for key, default_value in video_parameters:
        processing_parameters[key] = processing_parameters.get(key, default_value)
    return {**json_metadata, "processingParameters": processing_parameters}


def process_metadata(
    input_type: str,
    content_array: List[str],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    json_metadata_array: Union[List[dict], None],
):
    """
    Process the metadata of each asset
    """
    json_metadata_array = (
        [{}] * len(content_array) if json_metadata_array is None else json_metadata_array
    )
    if input_type in ("FRAME", "VIDEO"):
        should_use_native_video = json_content_array is None
        json_metadata_array = [
            add_video_parameters(json_metadata, should_use_native_video)
            for json_metadata in json_metadata_array
        ]
    return list(map(dumps, json_metadata_array))


def get_request_to_execute(
    input_type: str,
    json_metadata_array: Union[List[dict], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    mime_type: Union[str, None],
) -> Tuple[str, Optional[str]]:
    """
    Selects the right query to run versus the data given
    """
    if json_content_array is not None:
        return GQL_APPEND_MANY_TO_DATASET, None
    if input_type not in ("FRAME", "VIDEO"):
        if input_type == "IMAGE" and mime_type == "image/tiff":
            return GQL_APPEND_MANY_FRAMES_TO_DATASET, "GEO_SATELLITE"
        return GQL_APPEND_MANY_TO_DATASET, None
    if json_metadata_array is None:
        return GQL_APPEND_MANY_TO_DATASET, None
    if (
        isinstance(json_metadata_array, list)
        and len(json_metadata_array) > 0
        and not json_metadata_array[0]
        .get("processingParameters", {})
        .get("shouldUseNativeVideo", True)
    ):
        return GQL_APPEND_MANY_FRAMES_TO_DATASET, "VIDEO"
    return GQL_APPEND_MANY_TO_DATASET, None


# pylint: disable=too-many-arguments, too-many-locals
def process_append_many_to_dataset_parameters(
    auth: KiliAuth,
    input_type: str,
    content_array: Union[List[str], None],
    external_id_array: Union[List[str], None],
    is_honeypot_array: Union[List[str], None],
    status_array: Union[List[str], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    json_metadata_array: Union[List[dict], None],
):
    """
    Process arguments of the append_many_to_dataset method and return the data payload.
    """
    if content_array is None:
        assert (
            json_content_array
        ), "When content_array is not passed, you should pass json_content_array"
        content_array = [""] * len(json_content_array)
    if external_id_array is None:
        external_id_array = [uuid4().hex for _ in range(len(content_array))]
    is_honeypot_array_ = (
        [False] * len(content_array) if is_honeypot_array is None else is_honeypot_array
    )
    status_array = ["TODO"] * len(content_array) if not status_array else status_array
    formatted_json_metadata_array = process_metadata(
        input_type, content_array, json_content_array, json_metadata_array
    )

    mime_type = get_file_mimetype(content_array, json_content_array)  # type:ignore
    content_array = process_and_store_content(
        input_type, content_array, json_content_array, auth  # type:ignore X
    )
    formatted_json_content_array = process_json_content(
        input_type,
        content_array,  # type:ignore X
        json_content_array,  # type:ignore X
    )

    request, upload_type = get_request_to_execute(
        input_type, json_metadata_array, json_content_array, mime_type
    )
    properties = {
        "content_array": content_array,
        "external_id_array": external_id_array,
        "is_honeypot_array": is_honeypot_array_,
        "status_array": status_array,
        "json_content_array": formatted_json_content_array,
        "json_metadata_array": formatted_json_metadata_array,
    }
    return (properties, upload_type, request)



"""
Queries of asset mutations
"""

from kili.graphql.operations.asset.fragments import ASSET_FRAGMENT
from kili.mutations.project.fragments import PROJECT_FRAGMENT_ID

GQL_UPDATE_PROPERTIES_IN_ASSETS = f"""
mutation(
    $whereArray: [AssetWhere!]!
    $dataArray: [AssetData!]!
) {{
  data: updatePropertiesInAssets(
    where: $whereArray,
    data: $dataArray
  ) {{
    {ASSET_FRAGMENT}
  }}
}}
"""

GQL_DELETE_MANY_FROM_DATASET = f"""
mutation($where: AssetWhere!) {{
  data: deleteManyFromDataset(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW = f"""
mutation($where: AssetWhere!) {{
  data: addAllLabeledAssetsToReview(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_SEND_BACK_ASSETS_TO_QUEUE = f"""
mutation($where: AssetWhere!) {{
  data: sendBackAssetsToQueue(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""



"""User queries."""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from kili.helpers import format_result, fragment_builder
from kili.queries.user.queries import GQL_USERS_COUNT, gql_users
from kili.types import User
from kili.utils.pagination import row_generator_from_paginated_calls


class QueriesUser:
    """Set of User queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @typechecked
    def users(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: List[str] = ["email", "id", "firstname", "lastname"],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of users given a set of criteria

        Args:
            api_key: Query an user by its API Key
            email: Email of the user
            organization_id: Identifier of the user's organization
            fields: All the fields to request among the possible fields for the users.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#user) for all possible fields.
            first: Maximum number of users to return
            skip: Number of skipped users (they are ordered by creation date)
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the users is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            ```
            # List all users in my organization
            >>> organization = kili.organizations()
            >>> organization_id = organizations[0]['id]
            >>> kili.users(organization_id=organization_id)
            ```
        """

        count_args = {"organization_id": organization_id}
        disable_tqdm = disable_tqdm or as_generator or (api_key or email) is not None
        payload_query = {
            "where": {
                "apiKey": api_key,
                "email": email,
                "organization": {
                    "id": organization_id,
                },
            }
        }

        users_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_users,
            count_args,
            self._query_users,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return users_generator
        return list(users_generator)

    def _query_users(self, skip: int, first: int, payload: dict, fields: List[str]):
        payload.update({"skip": skip, "first": first})
        _gql_users = gql_users(fragment_builder(fields, User))
        result = self.auth.client.execute(_gql_users, payload)
        return format_result("data", result)

    @typechecked
    def count_users(self, organization_id: Optional[str] = None) -> int:
        """Get user count based on a set of constraints.

        Args:
            organization_id: Identifier of the user's organization

        Returns:
            The number of organizations with the parameters provided
        """
        variables = {
            "where": {
                "organization": {
                    "id": organization_id,
                }
            }
        }
        result = self.auth.client.execute(GQL_USERS_COUNT, variables)
        return format_result("data", result, int)



"""
Common code for the yolo exporter.
"""

import json
import os
from tempfile import TemporaryDirectory
from typing import Dict, List

from kili.orm import AnnotationFormat
from kili.services.export.format.base import BaseExporter


def _filter_out_autosave_labels(assets: List[Dict]):
    """
    Removes AUTOSAVE labels from exports

    Parameters
    ----------
    - assets: list of assets
    """
    clean_assets = []
    for asset in assets:
        labels = asset.get("labels", [])
        clean_labels = list(filter(lambda label: label["labelType"] != "AUTOSAVE", labels))
        if clean_labels:
            asset["labels"] = clean_labels
        clean_assets.append(asset)
    return clean_assets


def _format_json_response(label, label_format):
    """
    Format the label JSON response in the requested format
    """
    formatted_json_response = label.json_response(_format=label_format.lower())
    if label_format.lower() == AnnotationFormat.Simple:
        label["jsonResponse"] = formatted_json_response
    else:
        json_response = {}
        for key, value in formatted_json_response.items():
            if key.isdigit():
                json_response[int(key)] = value
                continue
            json_response[key] = value
        label["jsonResponse"] = json_response
    return label


def _process_assets(assets, label_format):
    """
    Format labels in the requested format, and filter out autosave labels
    """
    assets_in_format = []
    for asset in assets:
        if "labels" in asset:
            labels_of_asset = []
            for label in asset["labels"]:
                clean_label = _format_json_response(label, label_format)
                labels_of_asset.append(clean_label)
            asset["labels"] = labels_of_asset
        if "latestLabel" in asset:
            label = asset["latestLabel"]
            if label is not None:
                clean_label = _format_json_response(label, label_format)
                asset["latestLabel"] = clean_label
        assets_in_format.append(asset)

    clean_assets = _filter_out_autosave_labels(assets_in_format)
    return clean_assets


class KiliExporter(BaseExporter):
    """
    Common code for Kili exporters.
    """

    def _save_assets_export(self, assets: List[Dict], output_filename: str):
        """
        Save the assets to a file and return the link to that file
        """
        self.logger.info("Exporting to kili format...")

        with TemporaryDirectory() as root_folder:
            base_folder = os.path.join(root_folder, self.project_id)
            os.makedirs(base_folder)
            if self.single_file:
                project_json = json.dumps(assets, sort_keys=True, indent=4)
                with open(os.path.join(base_folder, "data.json"), "wb") as output_file:
                    output_file.write(project_json.encode("utf-8"))
            else:
                labels_folder = os.path.join(base_folder, "labels")
                os.makedirs(labels_folder)
                for asset in assets:
                    external_id = asset["externalId"].replace(" ", "_")
                    asset_json = json.dumps(asset, sort_keys=True, indent=4)
                    with open(
                        os.path.join(labels_folder, f"{external_id}.json"), "wb"
                    ) as output_file:
                        output_file.write(asset_json.encode("utf-8"))
            self.create_readme_kili_file(root_folder)
            self.make_archive(root_folder, output_filename)

        self.logger.warning(output_filename)

    def process_and_save(self, assets: List[Dict], output_filename: str):
        """
        Extract formatted annotations from labels and save the json in the buckets.
        """
        clean_assets = _process_assets(assets, self.label_format)
        return self._save_assets_export(
            clean_assets,
            output_filename,
        )



"""From https://docs.kili-technology.com/reference/graphql-api#enums"""
from typing_extensions import Literal

DataIntegrationPlatform = Literal[
    "AWS",
    "AZURE",
    "GCP",
]


DataIntegrationStatus = Literal["CONNECTEDDISCONNECTEDCHECKING"]


DataIntegrationType = Literal["REMOTE_STORAGE"]


ExportType = Literal["LATESTNORMAL"]


GraphScale = Literal["DAYMONTHWEEK"]


GraphType = Literal["COUNT_BY_CATEGORIESNUMBER_OF_LABELSHighestRoleInProject"]


HighestRoleInProject = Literal["NAADMINTEAM_MANAGERREVIEWERLABELER"]


HubspotSubscriptionStatus = Literal[
    "SUBSCRIBED",
    "UNSUBSCRIBED",
]


InputType = Literal[
    "AUDIO",
    "FRAME",
    "IMAGE",
    "NA",
    "PDF",
    "TEXT",
    "TIME_SERIES",
    "VIDEO",
    "URL",
]


InvitationStatus = Literal[
    "ACCEPTED",
    "DECLINED",
    "PENDING",
]


IssueStatus = Literal[
    "OPEN",
    "SOLVED",
]


IssueType = Literal[
    "ISSUE",
    "QUESTION",
]


LabelFormat = Literal[
    "RAW",
    "SIMPLE",
    "YOLO_V4",
    "YOLO_V5",
]


LabelType = Literal[
    "AUTOSAVE",
    "DEFAULT",
    "INFERENCE",
    "PREDICTION",
    "REVIEW",
]


LicenseType = Literal[
    "DISCOVERY",
    "FREEMIUM",
    "PAID",
]


LockType = Literal[
    "REVIEW",
    "DEFAULT",
]


NotificationStatus = Literal[
    "FAILURE",
    "PENDING",
    "SUCCESS",
]


OrganizationRole = Literal[
    "ADMIN",
    "USER",
]


ProjectRole = Literal[
    "ADMIN",
    "TEAM_MANAGER",
    "REVIEWER",
    "LABELER",
]


ProjectType = Literal[
    "IMAGE_CLASSIFICATION_SINGLE",
    "IMAGE_CLASSIFICATION_MULTI",
    "IMAGE_OBJECT_DETECTION_RECTANGLE",
    "IMAGE_OBJECT_DETECTION_POLYGON",
    "IMAGE_OBJECT_DETECTION_SEMANTIC",
    "IMAGE_POSE_ESTIMATION",
    "OCR",
    "PDF_CLASSIFICATION_SINGLE",
    "PDF_CLASSIFICATION_MULTI",
    "PDF_OBJECT_DETECTION_RECTANGLE",
    "PDF_NAMED_ENTITY_RECOGNITION",
    "SPEECH_TO_TEXT",
    "TEXT_CLASSIFICATION_SINGLE",
    "TEXT_CLASSIFICATION_MULTI",
    "TEXT_TRANSCRIPTION",
    "TEXT_NER",
    "TIME_SERIES",
    "VIDEO_CLASSIFICATION_SINGLE",
    "VIDEO_OBJECT_DETECTION",
    "VIDEO_FRAME_CLASSIFICATION",
    "VIDEO_FRAME_OBJECT_TRACKING",
]


Right = Literal[
    "CAN_ACCESS_SMART_TOOLS",
    "CAN_LABEL",
    "CREATE_AUDIO",
    "CREATE_FRAME",
    "CREATE_IMAGE",
    "CREATE_PDF",
    "CREATE_TEXT",
    "CREATE_TIMESERIES",
    "CREATE_VIDEO",
    "MAKE_PUBLIC_PROJECT",
    "SEE_LICENSE_BANNER",
    "SEE_UPGRADE_BUTTON",
    "UPLOAD_CLOUD_DATA",
    "USE_API",
    "USE_API_PRIORITY",
]


SplitOption = Literal[
    "MERGED_FOLDER",
    "SPLITTED_FOLDER",
]


Status = Literal[
    "TODO",
    "ONGOING",
    "LABELED",
    "REVIEWED",
    "TO_REVIEW",
]


UploadType = Literal[
    "VIDEO",
    "GEO_SATELLITE",
]


WarningTypes = Literal[
    "DUPLICATED_ID",
    "EMPTY_ARRAY",
    "FAILED_ASSET_CREATION",
    "FFMPEG_NOT_INSTALLED",
    "INPUT_UNDEFINED",
    "MAX_INPUT_SIZE",
    "MAX_PROJECT_SIZE",
    "MISSING_ID",
    "NON_INTEGER_FPS",
    "UNSUPPORTED_TYPE",
    "UNSPECIFIED_ERROR",
    "WRONG_CSV_FORMAT",
]



"""
Handles the Yolo export with the split layout
"""

import os
from tempfile import TemporaryDirectory
from typing import Dict, List

from ...exceptions import NoCompatibleJobError
from ...format.yolo.common import YoloExporter, get_category_full_name
from ...types import JobCategory


class YoloSplitExporter(YoloExporter):
    """
    Handles the Yolo format export into split folders
    """

    def process_and_save(self, assets: List[Dict], output_filename: str) -> None:
        self.logger.info("Exporting to yolo format split...")

        json_interface, ml_task, tool = self.get_project_and_init()
        categories_by_job = self._get_categories_by_job(json_interface, ml_task, tool)
        if not categories_by_job:
            raise NoCompatibleJobError(
                f"Error: There is no job in project {self.project_id} "
                f"that can be converted to the {self.label_format} format."
            )

        with TemporaryDirectory() as root_folder:
            images_folder = os.path.join(root_folder, self.project_id, "images")
            os.makedirs(images_folder)
            self._write_jobs_labels_into_split_folders(
                assets,
                categories_by_job,
                root_folder,
                images_folder,
            )
            self.create_readme_kili_file(root_folder)
            self.make_archive(root_folder, output_filename)

        self.logger.warning(output_filename)

    @classmethod
    def _get_categories_by_job(
        cls, json_interface: Dict, ml_task: str, tool: str
    ) -> Dict[str, Dict[str, JobCategory]]:
        """
        Return a dictionary of JobCategory instances by category full name and job id.
        """
        categories_by_job: Dict[str, Dict[str, JobCategory]] = {}
        for job_id, job in json_interface.get("jobs", {}).items():
            if (
                job.get("mlTask") != ml_task
                or tool not in job.get("tools", [])
                or job.get("isModel")
            ):
                continue
            categories: Dict[str, JobCategory] = {}
            for cat_id, category in enumerate(job.get("content", {}).get("categories", {})):
                categories[get_category_full_name(job_id, category)] = JobCategory(
                    category_name=category, id=cat_id, job_id=job_id
                )
            categories_by_job[job_id] = categories
        return categories_by_job

    def _write_jobs_labels_into_split_folders(
        self,
        assets: List[Dict],
        categories_by_job: Dict[str, Dict[str, JobCategory]],
        root_folder: str,
        images_folder: str,
    ) -> None:
        """
        Write assets into split folders.
        """
        for job_id, category_ids in categories_by_job.items():

            base_folder = os.path.join(root_folder, self.project_id, job_id)
            labels_folder = os.path.join(base_folder, "labels")
            os.makedirs(labels_folder)

            self.write_labels_into_single_folder(
                assets,
                category_ids,
                labels_folder,
                images_folder,
                base_folder,
            )



"""
Queries of lock queries
"""


def gql_locks(fragment):
    """
    Return the GraphQL locks query
    """
    return f"""
query($where: LockWhere!, $first: PageSize!, $skip: Int!) {{
  data: locks(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_LOCKS_COUNT = """
query($where: LockWhere!) {
  data: countLocks(where: $where)
}
"""



"""
Service exceptions
"""


class NotEnoughArgumentsSpecifiedError(ValueError):
    """
    Raised when there are not enough arguments specified in a service.
    """


class TooManyArgumentsSpecifiedError(ValueError):
    """
    Raised when there are too many arguments specified in a service.
    """



"""
Functions to export a project to YOLOv4, v5 or v7 format
"""
from typing import Type

from kili.services.export.types import SplitOption

from ..base import BaseExporter, BaseExporterSelector
from .merge import YoloMergeExporter
from .split import YoloSplitExporter


class YoloExporterSelector(BaseExporterSelector):
    # pylint: disable=too-few-public-methods

    """
    Formatter to export to YOLOv4, v5 or v7
    """

    @staticmethod
    def select_exporter_class(
        split_param: SplitOption,
    ) -> Type[BaseExporter]:
        """
        Export a project to YOLO v4 or v5 format
        """
        if split_param == "split":
            return YoloSplitExporter
        return YoloMergeExporter



"""
Queries of comment mutations
"""

from .fragments import COMMENT_FRAGMENT

GQL_APPEND_TO_COMMENTS = f"""
mutation(
    $data: AppendToCommentsData!
    $where: IssueWhere!
) {{
  data: appendToComments(
    data: $data
    where: $where
  ) {{
    {COMMENT_FRAGMENT}
  }}
}}
"""



"""CLI's project describe command"""

from typing import Dict, List, Optional, cast

import click
from tabulate import tabulate

from kili.cli.common_args import Arguments, Options
from kili.cli.helpers import get_kili_client
from kili.exceptions import NotFound
from kili.queries.project.helpers import get_project_metadata, get_project_metrics


@click.command(name="describe")
@Arguments.project_id
@Options.api_key
@Options.endpoint
def describe_project(api_key: Optional[str], endpoint: Optional[str], project_id: str):
    """Show project description and analytics.
    \b
    !!! Examples
        ```
        kili project describe --project-id <project_id>
        ```
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)
    projects: List[Dict] = []
    try:
        projects = cast(
            List[Dict],
            kili.projects(
                project_id=project_id,
                fields=[
                    "title",
                    "id",
                    "description",
                    "numberOfAssets",
                    "numberOfRemainingAssets",
                    "numberOfReviewedAssets",
                    "numberOfSkippedAssets",
                    "honeypotMark",
                    "consensusMark",
                    "numberOfOpenIssues",
                    "numberOfSolvedIssues",
                    "numberOfOpenQuestions",
                    "numberOfSolvedQuestions",
                ],
                disable_tqdm=True,
            ),
        )
    except:
        # pylint: disable=raise-missing-from
        raise NotFound(f"project ID: {project_id}")
    metadata = get_project_metadata(projects[0], kili.auth.client.endpoint)
    dataset_metrics, quality_metrics = get_project_metrics(projects[0])

    print(tabulate(metadata, tablefmt="plain"), end="\n" * 2)
    print("Dataset KPIs", end="\n" + "-" * len("Dataset KPIs") + "\n")
    print(tabulate(dataset_metrics, tablefmt="plain"), end="\n" * 2)
    print("Quality KPIs", end="\n" + "-" * len("Quality KPIs") + "\n")
    print(tabulate(quality_metrics, tablefmt="plain"))



"""
Assets related mutations
"""
from kili.graphql.operations.asset.fragments import PROJECT_FRAGMENT_ID

GQL_APPEND_MANY_TO_DATASET = f"""
mutation(
    $data: AppendManyToDatasetData!,
    $where: ProjectWhere!
  ) {{
  data: appendManyToDataset(
    data: $data,
    where: $where
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_APPEND_MANY_FRAMES_TO_DATASET = f"""
mutation(
    $data: AppendManyFramesToDatasetAsynchronouslyData!,
    $where: ProjectWhere!
  ) {{
  data: appendManyFramesToDatasetAsynchronously(
    data: $data,
    where: $where
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""



"""
Queries and organization queries
"""


def gql_organizations(fragment):
    """
    Return the GraphQL organizations query
    """
    return f"""
query($where: OrganizationWhere!, $first: PageSize!, $skip: Int!) {{
  data: organizations(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_ORGANIZATIONS_COUNT = """
query($where: OrganizationWhere!) {
  data: countOrganizations(where: $where)
}
"""

GQL_ORGANIZATION_METRICS = """
query($where: OrganizationMetricsWhere!) {
  data: organizationMetrics(where: $where) {
    numberOfAnnotations
    numberOfHours
    numberOfLabeledAssets
  }
}
"""



"""
This script permits to initialize the Kili Python SDK client.
"""
import os

from kili.authentication import KiliAuth
from kili.exceptions import AuthenticationFailed, NotFound
from kili.graphql_client import GraphQLClientName
from kili.mutations.api_key import MutationsApiKey
from kili.mutations.asset import MutationsAsset
from kili.mutations.issue import MutationsIssue
from kili.mutations.label import MutationsLabel
from kili.mutations.notification import MutationsNotification
from kili.mutations.organization import MutationsOrganization
from kili.mutations.project import MutationsProject
from kili.mutations.project_version import MutationsProjectVersion
from kili.mutations.user import MutationsUser
from kili.project import Project
from kili.queries.api_key import QueriesApiKey
from kili.queries.asset import QueriesAsset
from kili.queries.issue import QueriesIssue
from kili.queries.label import QueriesLabel
from kili.queries.lock import QueriesLock
from kili.queries.notification import QueriesNotification
from kili.queries.organization import QueriesOrganization
from kili.queries.project import QueriesProject
from kili.queries.project_user import QueriesProjectUser
from kili.queries.project_version import QueriesProjectVersion
from kili.queries.user import QueriesUser
from kili.services.types import ProjectId
from kili.subscriptions.label import SubscriptionsLabel


class Kili(  # pylint: disable=too-many-ancestors
    MutationsApiKey,
    MutationsAsset,
    MutationsIssue,
    MutationsLabel,
    MutationsNotification,
    MutationsOrganization,
    MutationsProject,
    MutationsProjectVersion,
    MutationsUser,
    QueriesApiKey,
    QueriesAsset,
    QueriesIssue,
    QueriesLabel,
    QueriesLock,
    QueriesOrganization,
    QueriesNotification,
    QueriesProject,
    QueriesProjectUser,
    QueriesProjectVersion,
    QueriesUser,
    SubscriptionsLabel,
):
    """
    Kili Client.
    """

    def __init__(
        self, api_key=None, api_endpoint=None, verify=True, client_name=GraphQLClientName.SDK
    ):
        """
        Args:
            api_key: User API key generated
                from https://cloud.kili-technology.com/label/my-account/api-key
                Default to  KILI_API_KEY environment variable).
                If not passed, requires the KILI_API_KEY environment variable to be set.
            api_endpoint: Recipient of the HTTP operation
                Default to  KILI_API_ENDPOINT environment variable).
                If not passed, default to Kili SaaS:
                'https://cloud.kili-technology.com/api/label/v2/graphql'
            verify: Verify certificate. Set to False on local deployment without SSL.
            client_name: For internal use only.
                Define the name of the graphQL client whith which graphQL calls will be sent.

        Returns:
            Object container your API session

        Examples:
            list:
                - your assets with: `kili.assets()`
                - your labels with: `kili.labels()`
                - your projects with: `kili.projects()`
        """
        if api_key is None:
            api_key = os.getenv("KILI_API_KEY")
        if api_endpoint is None:
            api_endpoint = os.getenv(
                "KILI_API_ENDPOINT",
                "https://cloud.kili-technology.com/api/label/v2/graphql",
            )

        if api_key is None:
            raise AuthenticationFailed(api_key, api_endpoint)
        try:
            self.auth = KiliAuth(
                api_key=api_key,
                api_endpoint=api_endpoint,
                client_name=client_name,
                verify=verify,
            )
            super().__init__(self.auth)
        except Exception as exception:  # pylint: disable=W0703
            exception_str = str(exception)
            if "b'Unauthorized'" in exception_str:
                raise AuthenticationFailed(api_key, api_endpoint) from exception
            raise exception

    def get_project(self, project_id: str) -> Project:
        """Return a project object corresponding to the project_id given.
        The returned project object inherit from many methods for project management

        Args:
            project_id: id of the project to return

        raise:
            NotFound if the given `project_id` does not correspond to an existing project
        """
        projects_response = self.projects(
            project_id=project_id, disable_tqdm=True, fields=["inputType", "title"]
        )
        projects_response = list(projects_response)

        if len(projects_response) == 0:
            raise NotFound(str(project_id))
        project_fields = projects_response[0]
        title = project_fields["title"]
        input_type = project_fields["inputType"]

        return Project(
            client=self, project_id=ProjectId(project_id), input_type=input_type, title=title
        )



"""
Label import-specific errors
"""


class LabelParsingError(ValueError):
    """
    Raised when there is a failure to parse a label
    """


class MissingMetadataError(ValueError):
    """
    Raised when a metadata file is missing
    """


class MissingTargetJobError(ValueError):
    """
    Raised when the target job is missing
    """


class MissingExternalIdError(ValueError):
    """
    Raised when the external id of the labels are missing
    """



"""Service for importing objects into kili"""

from typing import Dict, List

from kili.authentication import KiliAuth
from kili.constants import NO_ACCESS_RIGHT
from kili.queries.project import QueriesProject
from kili.services.asset_import.image import ImageDataImporter
from kili.services.asset_import.legacy import LegacyDataImporter
from kili.services.asset_import.pdf import PdfDataImporter
from kili.services.asset_import.text import TextDataImporter
from kili.services.asset_import.video import VideoDataImporter

from .base import LoggerParams, ProcessingParams, ProjectParams

importer_by_type = {
    "PDF": PdfDataImporter,
    "IMAGE": ImageDataImporter,
    "TEXT": TextDataImporter,
    "VIDEO": VideoDataImporter,
    "FRAMES": VideoDataImporter,
}


def import_assets(
    auth: KiliAuth,
    project_id: str,
    assets: List[Dict],
    raise_error=True,
    disable_tqdm=False,
):
    """
    import the selected assets into the specified project
    """
    kili = QueriesProject(auth)
    projects = list(kili.projects(project_id, disable_tqdm=True))
    assert len(projects) == 1, NO_ACCESS_RIGHT
    input_type = projects[0]["inputType"]

    project_params = ProjectParams(project_id=project_id, input_type=input_type)
    processing_params = ProcessingParams(raise_error=raise_error)
    logger_params = LoggerParams(disable_tqdm=disable_tqdm)
    importer_params = (auth, project_params, processing_params, logger_params)

    asset_importer = importer_by_type.get(input_type, LegacyDataImporter)(*importer_params)
    result = asset_importer.import_assets(assets=assets)  # type:ignore X
    return result



"""
Constants for the asset_improt services
"""

from uuid import uuid4

from .types import AssetLike

ASSET_FIELDS_DEFAULT_VALUE = AssetLike(
    content="",
    json_content="",
    external_id=uuid4().hex,
    status="TODO",
    json_metadata="{}",
    is_honeypot=False,
)

mime_extensions = {
    "Audio": "audio/x-flac,audio/mpeg,video/mp4",
    "Csv": "text/csv",
    "Video": "video/mp4,video/x-matroska,video/3gpp,video/x-msvideo,"
    "video/x-m4v,video/quicktime,video/webm",
    "Image": "image/jpeg,image/png,image/bmp,image/gif,image/webp,image/x-icon,"
    "image/tiff,image/vnd.microsoft.icon,image/svg+xml,image/avif,image/apng",
    "Pdf": "application/pdf",
    "Text": "text/plain",
    "TimeSeries": "text/csv",
}

project_compatible_mimetypes = {
    "AUDIO": mime_extensions["Audio"],
    "FRAME": mime_extensions["Video"],
    "IMAGE": mime_extensions["Image"],
    "NA": "",
    "PDF": mime_extensions["Pdf"],
    "TEXT": mime_extensions["Text"],
    "TIME_SERIES": mime_extensions["Csv"],
    "URL": "",
    "VIDEO": mime_extensions["Video"],
    "VIDEO_LEGACY": mime_extensions["Csv"],
}

IMPORT_BATCH_SIZE = 10
FRAME_IMPORT_BATCH_SIZE = 1

MB_SIZE = 1024**2
LARGE_IMAGE_THRESHOLD_SIZE = 30 * MB_SIZE



"""Notification queries."""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import format_result, fragment_builder
from ...types import Notification
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import GQL_NOTIFICATIONS_COUNT, gql_notifications


class QueriesNotification:
    """Set of Notification queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @typechecked
    def notifications(
        self,
        fields: List[str] = [
            "createdAt",
            "hasBeenSeen",
            "id",
            "message",
            "status",
            "userID",
        ],
        first: Optional[int] = None,
        has_been_seen: Optional[bool] = None,
        notification_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of notifications respecting a set of criteria.

        Args:
            fields: All the fields to request among the possible fields for the notifications
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#notification) for all possible fields.
            first: Number of notifications to query
            has_been_seen: If the notifications returned should have been seen.
            notification_id: If given, will return the notification which has this id
            skip: Number of notifications to skip (they are ordered by their date of creation,
                first to last).
            user_id: If given, returns the notifications of a specific user
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the notifications is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """

        count_args = {"has_been_seen": has_been_seen, "user_id": user_id}
        disable_tqdm = disable_tqdm or as_generator or notification_id is not None
        payload_query = {
            "where": {
                "id": notification_id,
                "user": {
                    "id": user_id,
                },
                "hasBeenSeen": has_been_seen,
            },
        }
        notifications_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_notifications,
            count_args,
            self._query_notifications,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return notifications_generator
        return list(notifications_generator)

    def _query_notifications(self, skip: int, first: int, payload: dict, fields: List[str]):
        payload.update({"skip": skip, "first": first})
        _gql_notifications = gql_notifications(fragment_builder(fields, Notification))
        result = self.auth.client.execute(_gql_notifications, payload)
        return format_result("data", result)

    @typechecked
    def count_notifications(
        self, has_been_seen: Optional[bool] = None, user_id: Optional[str] = None
    ) -> int:
        """Count the number of notifications.

        Args:
            has_been_seen: Filter on notifications that have been seen.
            user_id: Filter on the notifications of a specific user

        Returns:
            The number of notifications with the parameters provided
        """
        variables = {
            "where": {
                "user": {
                    "id": user_id,
                },
                "hasBeenSeen": has_been_seen,
            },
        }
        result = self.auth.client.execute(GQL_NOTIFICATIONS_COUNT, variables)
        return format_result("data", result, int)



"""
Common and generic functions to import files into a project
"""
import mimetypes
import os
from concurrent.futures import ThreadPoolExecutor
from json import dumps
from typing import Callable, List, NamedTuple

from kili.authentication import KiliAuth
from kili.graphql.operations.asset.mutations import (
    GQL_APPEND_MANY_FRAMES_TO_DATASET,
    GQL_APPEND_MANY_TO_DATASET,
)
from kili.helpers import T, format_result, is_url
from kili.orm import Asset
from kili.queries.asset import QueriesAsset
from kili.services.asset_import.constants import (
    ASSET_FIELDS_DEFAULT_VALUE,
    IMPORT_BATCH_SIZE,
    project_compatible_mimetypes,
)
from kili.services.asset_import.exceptions import ImportValidationError, MimeTypeError
from kili.services.asset_import.types import AssetLike, KiliResolverAsset
from kili.utils import bucket, pagination
from kili.utils.tqdm import tqdm


class BatchParams(NamedTuple):
    """
    Contains all parameters related to the batch to import
    """

    is_asynchronous: bool
    is_hosted: bool


class ProcessingParams(NamedTuple):
    """
    Contains all parameters related to the assets processing
    """

    raise_error: bool


class ProjectParams(NamedTuple):
    """
    Contains all parameters related to the batch to import
    """

    project_id: str
    input_type: str


class LoggerParams(NamedTuple):
    """
    Contains all parameters related to logging
    """

    disable_tqdm: bool


class BaseBatchImporter:  # pylint: disable=too-few-public-methods
    """
    Base class for BatchImporters
    """

    def __init__(
        self, auth: KiliAuth, project_params: ProjectParams, batch_params: BatchParams, pbar: tqdm
    ):
        self.auth = auth
        self.project_id = project_params.project_id
        self.input_type = project_params.input_type
        self.is_hosted = batch_params.is_hosted
        self.is_asynchronous = batch_params.is_asynchronous
        self.pbar = pbar

    @pagination.api_throttle
    def import_batch(self, assets: List[AssetLike]):
        """
        Base actions to import a batch of asset
        """
        assets = self.loop_on_batch(self.stringify_metadata)(assets)
        assets_ = self.loop_on_batch(self.fill_empty_fields)(assets)
        result_batch = self.import_to_kili(assets_)
        self.pbar.update(n=len(assets))
        return result_batch

    @staticmethod
    def stringify_metadata(asset: AssetLike) -> AssetLike:
        """
        Stringify the metadata
        """
        json_metadata = asset.get("json_metadata", {})
        if not isinstance(json_metadata, str):
            json_metadata = dumps(json_metadata)
        return {**asset, "json_metadata": json_metadata}

    @staticmethod
    def fill_empty_fields(asset: AssetLike):
        """
        fill empty fields with their default value
        """
        field_names = ASSET_FIELDS_DEFAULT_VALUE.keys()
        return KiliResolverAsset(
            **{field: asset.get(field, ASSET_FIELDS_DEFAULT_VALUE[field]) for field in field_names}
        )

    @staticmethod
    def loop_on_batch(func: Callable[[AssetLike], T]) -> Callable[[List[AssetLike]], List[T]]:
        """
        Apply a function, that takes a single asset as input, on the whole batch
        """

        def loop_func(assets: List[AssetLike]):
            return [func(asset) for asset in assets]

        return loop_func

    def _async_import_to_kili(self, assets: List[KiliResolverAsset]):
        """
        Import assets with asynchronous resolver.
        """
        upload_type = "GEO_SATELLITE" if self.input_type == "IMAGE" else "VIDEO"
        payload = {
            "data": {
                "contentArray": [asset["content"] for asset in assets],
                "externalIDArray": [asset["external_id"] for asset in assets],
                "jsonMetadataArray": [asset["json_metadata"] for asset in assets],
                "uploadType": upload_type,
            },
            "where": {"id": self.project_id},
        }
        results = self.auth.client.execute(GQL_APPEND_MANY_FRAMES_TO_DATASET, payload)
        return format_result("data", results, Asset)

    def _sync_import_to_kili(self, assets: List[KiliResolverAsset]):
        """
        Import assets with synchronous resolver
        """
        payload = {
            "data": {
                "contentArray": [asset["content"] for asset in assets],
                "externalIDArray": [asset["external_id"] for asset in assets],
                "isHoneypotArray": [asset["is_honeypot"] for asset in assets],
                "statusArray": [asset["status"] for asset in assets],
                "jsonContentArray": [asset["json_content"] for asset in assets],
                "jsonMetadataArray": [asset["json_metadata"] for asset in assets],
            },
            "where": {"id": self.project_id},
        }
        results = self.auth.client.execute(GQL_APPEND_MANY_TO_DATASET, payload)
        return format_result("data", results, Asset)

    def import_to_kili(self, assets: List[KiliResolverAsset]):
        """
        Import assets to Kili with the right resolver
        """
        if self.is_asynchronous:
            return self._async_import_to_kili(assets)
        return self._sync_import_to_kili(assets)


class ContentBatchImporter(BaseBatchImporter):
    """
    Class defining the methods to import of a batch of assets with content
    """

    @pagination.api_throttle
    def import_batch(self, assets: List[AssetLike]):
        """
        Method to import a batch of asset with content
        """
        if not self.is_hosted:
            assets = self.upload_local_content_to_bucket(assets)
        return super().import_batch(assets)

    def upload_local_content_to_bucket(self, assets: List[AssetLike]):
        """
        Upload local content to a bucket
        """
        signed_urls = bucket.request_signed_urls(self.auth, len(assets))
        data_array = []
        content_type_array = []
        for asset in assets:
            path = asset.get("content")
            assert path
            with open(path, "rb") as file:
                data_array.append(file.read())
            content_type, _ = mimetypes.guess_type(path)
            content_type_array.append(content_type)
        with ThreadPoolExecutor() as threads:
            url_gen = threads.map(
                bucket.upload_data_via_rest, signed_urls, data_array, content_type_array
            )
        return [AssetLike(**{**asset, "content": url}) for asset, url in zip(assets, url_gen)]


class JsonContentBatchImporter(BaseBatchImporter):
    """
    Class defining the import methods for a batch of assets twith json_content
    """

    @staticmethod
    def stringify_json_content(asset: AssetLike):
        """
        Stringify the json content if not a str.
        """
        json_content = asset.get("json_content", {})
        if not isinstance(json_content, str):
            json_content = dumps(json_content)
        return AssetLike(**{**asset, "json_content": json_content})

    def upload_json_content_to_bucket(self, assets: List[AssetLike]):
        """
        Upload the json_contents to a bucket with signed urls
        """
        signed_urls = bucket.request_signed_urls(self.auth, len(assets))
        json_content_array = [asset.get("json_content") for asset in assets]
        with ThreadPoolExecutor() as threads:
            url_gen = threads.map(
                bucket.upload_data_via_rest,
                signed_urls,
                json_content_array,
                ["text/plain"] * len(assets),
            )
        return [AssetLike(**{**asset, "json_content": url}) for asset, url in zip(assets, url_gen)]

    @pagination.api_throttle
    def import_batch(self, assets: List[AssetLike]):
        """
        Method to import a batch of asset with json content
        """
        assets = self.loop_on_batch(self.stringify_json_content)(assets)
        assets = self.upload_json_content_to_bucket(assets)
        return super().import_batch(assets)


class BaseAssetImporter:
    """
    Base class for DataImporters classes
    """

    def __init__(
        self,
        auth: KiliAuth,
        project_params: ProjectParams,
        processing_params: ProcessingParams,
        logger_params: LoggerParams,
    ):
        self.auth = auth
        self.project_params = project_params
        self.raise_error = processing_params.raise_error
        self.pbar = tqdm(disable=logger_params.disable_tqdm)

    @staticmethod
    def is_hosted_content(assets: List[AssetLike]):
        """
        Determine if the assets to upload are from local files or hosted data
        Raise an error if a mix of both
        """
        contents = [asset.get("content") for asset in assets]
        if all(is_url(content) for content in contents):
            return True
        if any(is_url(content) for content in contents):
            raise ImportValidationError(
                """
                Cannot upload hosted data and local files at the same time.
                Please separate the assets into 2 calls
                """
            )
        return False

    def filter_local_assets(self, assets: List[AssetLike], raise_error: bool):
        """
        Filter out local files that cannot be imported.
        Return an error at the first file that cannot be imported if raise_error is True
        """
        filtered_assets = []
        for asset in assets:
            path = asset.get("content")
            assert path
            try:
                self.check_mime_type_compatibility(path)
                filtered_assets.append(asset)
            except (FileNotFoundError, MimeTypeError) as err:
                if raise_error:
                    raise err
        if len(filtered_assets) == 0:
            raise ImportValidationError(
                """
                No files to upload.
                Check that the paths exist and file types are compatible with the project
                """
            )
        return filtered_assets

    def check_mime_type_compatibility(self, path: str):
        """
        Check that the mimetype of a local file is compatible with the project input type.

        Raise:
             FileNotFoundError: if path do not exists
             MimeTypeError: if the asset is not compatible with the project type
        """
        if not os.path.isfile(path):
            raise FileNotFoundError(f"file {path} does not exist")
        mime_type, _ = mimetypes.guess_type(path)
        if mime_type is None:
            raise MimeTypeError(f"The mime type of the asset {path} has not been found")

        input_type = self.project_params.input_type
        if mime_type not in project_compatible_mimetypes[input_type]:
            raise MimeTypeError(
                f"""
                File mime type for {path} is {mime_type} and does not correspond
                to the type of the project.
                File mime type should be one of {project_compatible_mimetypes[input_type]}
                """
            )
        return True

    def filter_duplicate_external_ids(self, assets):
        """Filter out assets whose external_id is already in the project."""
        if len(assets) == 0:
            raise ImportValidationError("No assets to import")
        assets_in_project = QueriesAsset(self.auth).assets(
            project_id=self.project_params.project_id, fields=["externalId"], disable_tqdm=True
        )
        external_ids_in_project = [asset["externalId"] for asset in assets_in_project]
        filtered_assets = [
            asset for asset in assets if asset.get("external_id") not in external_ids_in_project
        ]
        if len(filtered_assets) == 0:
            raise ImportValidationError(
                "No assets to import, all given external_ids already exist in the project"
            )
        return filtered_assets

    def import_assets_by_batch(
        self,
        assets: List[AssetLike],
        batch_importer: BaseBatchImporter,
        batch_size=IMPORT_BATCH_SIZE,
    ):
        """Split assets by batch and import them with a given batch importer."""
        batch_generator = pagination.batch_iterator_builder(assets, batch_size)
        self.pbar.total = len(assets)
        self.pbar.refresh()
        responses = []
        for batch_assets in batch_generator:
            responses.append(batch_importer.import_batch(batch_assets))
        return responses[-1]



"""
GraphQL Client
"""


import json
import os
import random
import string
import threading
import time
import urllib
from datetime import datetime
from enum import Enum

import websocket

from kili import __version__


class GraphQLClientName(Enum):  # pylint: disable=too-few-public-methods
    """GraphQL client name."""

    SDK = "python-sdk"
    CLI = "python-cli"


class GraphQLClient:
    """
    A simple GraphQL client
    """

    def __init__(
        self,
        endpoint,
        client_name: GraphQLClientName,
        session=None,
        verify=True,
    ):
        self.endpoint = endpoint
        self.client_name = client_name
        self.headername = None
        self.session = session
        self.token = None
        self.verify = verify

    def execute(self, query, variables=None):
        """
        Execute a query

        Args:
            query
            variables
        """
        return self._send(query, variables)

    def inject_token(self, token, headername="Authorization"):
        """Inject a token.

        Args:
            token
            headername:
        """
        self.token = token
        self.headername = headername

    def _send(self, query, variables):
        """
        Send the query

        Args:
            query
            variables
        """
        data = {"query": query, "variables": variables}
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "apollographql-client-name": self.client_name.value,
            "apollographql-client-version": __version__,
        }

        if self.token is not None:
            if self.headername:
                headers[self.headername] = f"{self.token}"
            else:
                raise ValueError("headername must be defined.")

        if self.session is not None:
            req = None
            try:
                try:
                    number_of_trials = int(os.getenv("KILI_SDK_TRIALS_NUMBER", "10"))
                except ValueError:
                    number_of_trials = 10
                for trial_number in range(number_of_trials):
                    self.session.verify = self.verify
                    req = self.session.post(
                        self.endpoint, json.dumps(data).encode("utf-8"), headers=headers
                    )
                    errors_in_response = "errors" in req.json()
                    if req.status_code == 200 and not errors_in_response:
                        break
                    if req.status_code == 401:
                        raise Exception("Invalid API KEY")
                    if trial_number == number_of_trials - 1 and errors_in_response:
                        break
                    time.sleep(1)
                return req.json()  # type:ignore X
            except Exception as exception:
                if req is not None:
                    raise Exception(req.content) from exception
                raise exception

        req = urllib.request.Request(  # type: ignore
            self.endpoint, json.dumps(data).encode("utf-8"), headers
        )
        try:
            with urllib.request.urlopen(req) as response:  # type:ignore
                str_json = response.read().decode("utf-8")
                return json.loads(str_json)
        except urllib.error.HTTPError as error:  # type:ignore
            print((error.read()))
            print("")
            raise error


GQL_WS_SUBPROTOCOL = "graphql-ws"


class SubscriptionGraphQLClient:
    """
    A simple GraphQL client that works over Websocket as the transport
    protocol, instead of HTTP.
    This follows the Apollo protocol.
    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self, url):
        self.ws_url = url
        self._id = None
        self._paused = False
        self._connect()
        self._subscription_running = False
        self._st_id = None
        self.failed_connection_attempts = 0

    def _connect(self):
        """
        Handles the connection
        """
        # pylint: disable=no-member
        self._conn = websocket.create_connection(
            self.ws_url, on_message=self._on_message, subprotocols=[GQL_WS_SUBPROTOCOL]
        )
        self._created_at = datetime.now()
        self._conn.on_message = self._on_message

    def _reconnect(self):
        """
        Handles the reconnection
        """
        self._connect()
        self._subscription_running = True
        dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        print(f"{dt_string} reconnected")
        self.failed_connection_attempts = 0

    def _on_message(self, message):
        """
        Handles messages

        Args:
            message : the message
        """
        data = json.loads(message)
        # skip keepalive messages
        if data["type"] != "ka":
            print(message)

    def _conn_init(self, headers=None, authorization=None):
        """
        Initializes the websocket connection

        Args:
            headers : Headers are necessary for Kili API v1
            authorization : Headers are necessary for Kili API v2
        """
        payload = {
            "type": "connection_init",
            "payload": {"headers": headers, "Authorization": authorization},
        }
        self._conn.send(json.dumps(payload))
        self._conn.recv()

    def _start(self, payload):
        """
        Handles start

        Args:
            payload
        """
        _id = gen_id()
        frame = {"id": _id, "type": "start", "payload": payload}
        self._conn.send(json.dumps(frame))
        return _id

    def _stop(self, _id):
        """
        Handles stop

        Args:
        - _id: connection id
        """
        payload = {"id": _id, "type": "stop"}
        self._conn.send(json.dumps(payload))
        return self._conn.recv()

    def query(self, query, variables=None, headers=None):
        """
        Sends a query

        Args:
            query
            variables
            headers
        """
        self._conn_init(headers)
        payload = {"headers": headers, "query": query, "variables": variables}
        _id = self._start(payload)
        res = self._conn.recv()
        self._stop(_id)
        return res

    def prepare_subscribe(self, query, variables, headers, callback, authorization):
        """
        Prepares a subscription

        Args:
            query
            variables
            headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        self._conn_init(headers, authorization)
        payload = {"headers": headers, "query": query, "variables": variables}
        _cc = self._on_message if not callback else callback
        _id = self._start(payload)
        self._id = _id
        return _cc, _id

    def subscribe(self, query, variables=None, headers=None, callback=None, authorization=None):
        """
        Subscribes

        Args:
            query
            variables
            headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        _cc, _id = self.prepare_subscribe(query, variables, headers, callback, authorization)

        def subs(_cc, _id):
            max_reconnections = 10
            self._subscription_running = True
            while (
                self._subscription_running and self.failed_connection_attempts < max_reconnections
            ):
                try:
                    response = json.loads(self._conn.recv())
                    if response["type"] == "error" or response["type"] == "complete":
                        print(response)
                        self._stop_subscribe(_id)
                        break
                    if response["type"] != "ka" and not self._paused:
                        _cc(_id, response)  # type:ignore
                    time.sleep(1)
                except (
                    websocket._exceptions.WebSocketConnectionClosedException  # pylint: disable=no-member,protected-access
                ) as error:
                    self.failed_connection_attempts += 1
                    dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                    error_message = str(error)
                    print(f"{dt_string} Connection closed error : {error_message}")
                    print(
                        "Will try to reconnect"
                        f" {max_reconnections - self.failed_connection_attempts} times..."
                    )
                    self._reconnect()
                    _cc, _id = self.prepare_subscribe(
                        query, variables, headers, callback, authorization
                    )
                    continue
            print(f"Did not reconnect successfully after {max_reconnections} attempts")

        self._st_id = threading.Thread(target=subs, args=(_cc, _id))
        self._st_id.start()
        return _id

    def _stop_subscribe(self, _id):
        self._subscription_running = False
        self._stop(_id)

    def close(self):
        """
        Handles close
        """
        self._conn.close()

    def pause(self):
        """
        Handles pause
        """
        self._paused = True

    def unpause(self):
        """
        Handles unpause
        """
        self._paused = False

    def get_lifetime(self):
        """
        Return the lifetime
        """
        return (datetime.now() - self._created_at).seconds

    def reset_timeout(self):
        """
        Resets the timeout
        """
        self._reconnect()


def gen_id(size=6, chars=string.ascii_letters + string.digits):
    """
    Generate random alphanumeric id

    Args:
        size: length of the id
        chars: chars used to generate the id
    """
    return "".join(random.choice(chars) for _ in range(size))



"""
Functions to export a project to YOLOv4 or v5 format, with a merged folder layout
"""
import os
from tempfile import TemporaryDirectory
from typing import Dict, List

from ...exceptions import NoCompatibleJobError
from ...format.yolo.common import YoloExporter, get_category_full_name
from ...types import JobCategory


class YoloMergeExporter(YoloExporter):
    """
    Handles the Yolo export to merged folders.
    """

    def process_and_save(self, assets: List[Dict], output_filename: str) -> None:
        self.logger.info("Exporting to yolo format merged...")
        json_interface, ml_task, tool = self.get_project_and_init()
        merged_categories_id = self._get_merged_categories(json_interface, ml_task, tool)
        if not merged_categories_id:
            raise NoCompatibleJobError(
                f"Error: There is no job in project {self.project_id} "
                f"that can be converted to the {self.label_format} format."
            )

        with TemporaryDirectory() as root_folder:
            base_folder = os.path.join(root_folder, self.project_id)
            images_folder = os.path.join(base_folder, "images")
            labels_folder = os.path.join(base_folder, "labels")
            os.makedirs(images_folder)
            os.makedirs(labels_folder)
            self.write_labels_into_single_folder(
                assets,
                merged_categories_id,
                labels_folder,
                images_folder,
                base_folder,
            )
            self.create_readme_kili_file(root_folder)
            self.make_archive(root_folder, output_filename)

        self.logger.warning(output_filename)

    @classmethod
    def _get_merged_categories(
        cls, json_interface: Dict, ml_task: str, tool: str
    ) -> Dict[str, JobCategory]:
        """
        Return a dictionary of JobCategory instances by category full name.
        """
        cat_number = 0
        merged_categories_id: Dict[str, JobCategory] = {}
        for job_id, job in json_interface.get("jobs", {}).items():
            if (
                job.get("mlTask") != ml_task
                or tool not in job.get("tools", [])
                or job.get("isModel")
            ):
                continue
            for category in job.get("content", {}).get("categories", {}):
                merged_categories_id[get_category_full_name(job_id, category)] = JobCategory(
                    category_name=category, id=cat_number, job_id=job_id
                )
                cat_number += 1

        return merged_categories_id



"""
Types used by the conversion service
"""
from typing import NamedTuple, NewType, Tuple, Union

from typing_extensions import Literal

InputType = Literal["TEXT", "IMAGE"]
AssetId = NewType("AssetId", str)
ProjectId = NewType("ProjectId", str)

LogLevel = Union[int, Literal["ERROR", "WARNING", "DEBUG", "INFO", "CRITICAL"]]


class JobCategory(NamedTuple):
    """
    Contains information for a category
    """

    category_name: str
    id: int
    job_id: str


YoloAnnotation = Tuple[int, float, float, float, float]
LabelType = Literal["AUTOSAVE", "DEFAULT", "PREDICTION", "REVIEW"]



"""Module for managing bucket's signed urls"""


from typing import Union
from urllib.parse import parse_qs, urlparse

import requests
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from kili.authentication import KiliAuth
from kili.graphql.operations.asset.queries import GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS

AZURE_STRING = "blob.core.windows.net"
GCP_STRING = "storage.googleapis.com"
GCP_STRING_PUBLIC = "storage.cloud.google.com"

MAX_NUMBER_SIGNED_URLS_TO_FETCH = 30


def request_signed_urls(auth: KiliAuth, size: int):
    """
    Get upload signed URLs
    Args:
        auth: Kili Auth
        project_id: the project Id
        size: the amount of upload signed URL to query
    """
    signed_urls = []
    while len(signed_urls) < size:
        nb_url_requested = min(MAX_NUMBER_SIGNED_URLS_TO_FETCH, size - len(signed_urls))
        payload = {
            "size": nb_url_requested,
        }
        urls_response = auth.client.execute(GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS, payload)
        signed_urls.extend(urls_response["data"]["urls"])
    return signed_urls


@retry(stop=stop_after_attempt(3), wait=wait_random(min=1, max=2))
def upload_data_via_rest(url_with_id: str, data: Union[str, bytes], content_type: str):
    """upload data in buckets' signed URL via REST
    Args:
        signed_urls: Bucket signed URLs to upload local files to
        path_array: a list of file paths, json or text to upload
        content_type: mimetype of the data. It will be infered if not given
    """
    headers = {"Content-type": content_type}
    url_to_use_for_upload = url_with_id.split("&id=")[0]
    if "blob.core.windows.net" in url_to_use_for_upload:
        headers["x-ms-blob-type"] = "BlockBlob"

    response = requests.put(url_to_use_for_upload, data=data, headers=headers, timeout=30)
    response.raise_for_status()
    return url_with_id


def clean_signed_url(url: str, endpoint: str):
    """
    return a cleaned sined url for frame upload
    """
    query = urlparse(url).query
    id_param = parse_qs(query)["id"][0]
    base_path = endpoint.replace("/graphql", "/files").replace("http://", "https://")
    return f"{base_path}?id={id_param}"



"""
Base class for all formatters and other utility classes.
"""

import logging
import os
import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict, List, NamedTuple, Optional, Type

from kili.orm import JobMLTask, JobTool
from kili.services.export.repository import SDKContentRepository
from kili.services.export.tools import fetch_assets
from kili.services.export.types import ExportType, LabelFormat, SplitOption
from kili.services.types import LogLevel


class ExportParams(NamedTuple):
    """
    Contains all parameters that change the result of the export
    """

    assets_ids: Optional[List[str]]
    export_type: ExportType
    project_id: str
    label_format: LabelFormat
    split_option: SplitOption
    single_file: bool
    output_file: str


class LoggerParams(NamedTuple):
    """
    Contains all parameters related to logging.
    """

    disable_tqdm: bool
    level: LogLevel


class ContentRepositoryParams(NamedTuple):
    """
    Contains all the parameters related to the content repository.
    """

    router_endpoint: str
    router_headers: Dict[str, str]


class BaseExporter(ABC):
    """
    Abstract class defining the interface for all exporters.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        project_id,
        export_type,
        label_format,
        single_file,
        disable_tqdm,
        kili,
        logger,
        content_repository,
    ):  # pylint: disable=too-many-arguments
        self.project_id = project_id
        self.export_type = export_type
        self.label_format = label_format
        self.single_file = single_file
        self.disable_tqdm = disable_tqdm
        self.kili = kili
        self.logger = logger
        self.content_repository = content_repository

    @abstractmethod
    def process_and_save(self, assets: List[Dict], output_filename: str) -> None:
        """
        Converts the asset and save them into an archive.
        """

    def make_archive(self, root_folder: str, output_filename: str) -> str:
        """
        Make the export archive
        """
        path_folder = os.path.join(root_folder, self.project_id)
        path_archive = shutil.make_archive(path_folder, "zip", path_folder)
        shutil.copy(path_archive, output_filename)
        return output_filename

    def get_project_and_init(self):
        """
        Get and validate the project
        """
        json_interface = self.kili.projects(
            project_id=self.project_id, fields=["jsonInterface"], disable_tqdm=True
        )[0]["jsonInterface"]

        ml_task = JobMLTask.ObjectDetection
        tool = JobTool.Rectangle

        return json_interface, ml_task, tool

    def create_readme_kili_file(self, root_folder: str) -> None:
        """
        Create a README.kili.txt file to give information about exported labels
        """
        readme_file_name = os.path.join(root_folder, self.project_id, "README.kili.txt")
        project_info = self.kili.projects(
            project_id=self.project_id, fields=["title", "id", "description"], disable_tqdm=True
        )[0]
        with open(readme_file_name, "wb") as fout:
            fout.write("Exported Labels from KILI\n=========================\n\n".encode())
            fout.write(f"- Project name: {project_info['title']}\n".encode())
            fout.write(f"- Project identifier: {self.project_id}\n".encode())
            fout.write(f"- Project description: {project_info['description']}\n".encode())
            fout.write(f'- Export date: {datetime.now().strftime("%Y%m%d-%H%M%S")}\n'.encode())
            fout.write(f"- Exported format: {self.label_format}\n".encode())
            fout.write(f"- Exported labels: {self.export_type}\n".encode())

    def export_project(
        self,
        kili,
        export_params: ExportParams,
        logger_params: LoggerParams,
    ) -> None:
        """
        Export a project to a json.
        Return the name of the exported archive file in the bucket.
        """
        assets = fetch_assets(
            kili,
            project_id=export_params.project_id,
            asset_ids=export_params.assets_ids,
            export_type=export_params.export_type,
            label_type_in=["DEFAULT", "REVIEW"],
            disable_tqdm=logger_params.disable_tqdm,
        )
        self.process_and_save(assets, export_params.output_file)


class BaseExporterSelector(ABC):
    # pylint: disable=too-few-public-methods

    """
    Abstract class defining a standard signature for all formatters
    """

    @staticmethod
    def get_logger_assets_and_content_repo(
        kili,
        export_params: ExportParams,
        logger_params: LoggerParams,
        content_repository_params: ContentRepositoryParams,
    ):
        """
        Fetches assets and gets right logger and content repository depending on parameters
        """
        logger = BaseExporterSelector.get_logger(logger_params.level)

        logger.info("Fetching assets ...")
        assets = fetch_assets(
            kili,
            project_id=export_params.project_id,
            asset_ids=export_params.assets_ids,
            export_type=export_params.export_type,
            label_type_in=["DEFAULT", "REVIEW"],
            disable_tqdm=logger_params.disable_tqdm,
        )
        content_repository = SDKContentRepository(
            content_repository_params.router_endpoint,
            content_repository_params.router_headers,
            verify_ssl=True,
        )
        return logger, assets, content_repository

    @staticmethod
    @abstractmethod
    def select_exporter_class(
        split_param: SplitOption,
    ) -> Type[BaseExporter]:
        """
        Return the right exporter class.
        """

    def select_exporter(
        self,
        kili,
        logger_params,
        export_params: ExportParams,
        content_repository_params: ContentRepositoryParams,
    ) -> BaseExporter:
        """
        Return the right exporter.
        """
        logger = BaseExporterSelector.get_logger(logger_params.level)

        logger.info("Fetching assets ...")
        content_repository = SDKContentRepository(
            content_repository_params.router_endpoint,
            content_repository_params.router_headers,
            verify_ssl=True,
        )
        exporter_class = self.select_exporter_class(export_params.split_option)
        return exporter_class(
            export_params.project_id,
            export_params.export_type,
            export_params.label_format,
            export_params.single_file,
            logger_params.disable_tqdm,
            kili,
            logger,
            content_repository,
        )

    @staticmethod
    def get_logger(level: LogLevel):
        """Gets the export logger"""
        logger = logging.getLogger("kili.services.export")
        logger.setLevel(level)
        if logger.hasHandlers():
            logger.handlers.clear()
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        return logger



"""User mutations."""

from typing import Any, Dict, Optional

from typeguard import typechecked

from kili.helpers import format_result
from kili.mutations.user.queries import (
    GQL_CREATE_USER,
    GQL_RESET_PASSWORD,
    GQL_UPDATE_PASSWORD,
    GQL_UPDATE_PROPERTIES_IN_USER,
)


class MutationsUser:
    """Set of User mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def create_user(
        self,
        email: str,
        password: str,
        organization_role: str,
        firstname: Optional[str] = None,
        lastname: Optional[str] = None,
    ):
        """Add a user to your organization.

        Args:
            email: Email of the new user, used as user's unique identifier.
            password: On the first sign in, he will use this password and be able to change it.
            organization_role: One of "ADMIN", "USER".
            firstname: First name of the new user.
            lastname: Last name of the new user.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            "data": {
                "email": email,
                "password": password,
                "organizationRole": organization_role,
            }
        }
        if firstname is not None:
            variables["data"]["firstname"] = firstname
        if lastname is not None:
            variables["data"]["lastname"] = lastname
        result = self.auth.client.execute(GQL_CREATE_USER, variables)
        return format_result("data", result)

    @typechecked
    def update_password(
        self, email: str, old_password: str, new_password_1: str, new_password_2: str
    ):
        """Allow to modify the password that you use to connect to Kili. \
        This resolver only works for on-premise installations without Auth0.

        Args:
            email : Email of the person whose password has to be updated.
            old_password : The old password
            new_password_1: The new password
            new_password_2: A confirmation field for the new password

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            "data": {
                "oldPassword": old_password,
                "newPassword1": new_password_1,
                "newPassword2": new_password_2,
            },
            "where": {"email": email},
        }
        result = self.auth.client.execute(GQL_UPDATE_PASSWORD, variables)
        return format_result("data", result)

    @typechecked
    def reset_password(self, email: str):
        """Reset password.

        This resolver only works for on-premise installations without Auth0,
        if your organization allows Kili to send emails.

        Args:
            email: Email of the person whose password has to be reset.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"where": {"email": email}}
        result = self.auth.client.execute(GQL_RESET_PASSWORD, variables)
        return format_result("data", result)

    @typechecked
    def update_properties_in_user(
        self,
        email: str,
        firstname: Optional[str] = None,
        lastname: Optional[str] = None,
        organization_id: Optional[str] = None,
        organization_role: Optional[str] = None,
        activated: Optional[bool] = None,
    ):
        """Update the properties of a user.

        Args:
            email: The email is the identifier of the user.
            firstname:Change the first name of the user.
            lastname: Change the last name of the user.
            organization_id: Change the organization the user is related to.
            organization_role: Change the role of the user.
                One of "ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER".
            activated: In case we want to deactivate a user, but keep it.


        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables: Dict[str, Any] = {
            "email": email,
        }
        if firstname is not None:
            variables["firstname"] = firstname
        if lastname is not None:
            variables["lastname"] = lastname
        if organization_id is not None:
            variables["organizationId"] = organization_id
        if organization_role is not None:
            variables["organizationRole"] = organization_role
        if activated is not None:
            variables["activated"] = activated
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_USER, variables)
        return format_result("data", result)



"""
Functions to export a project to Kili format
"""
from typing import Type

from kili.services.export.format.kili.common import KiliExporter
from kili.services.export.types import SplitOption

from ..base import BaseExporter, BaseExporterSelector


class KiliExporterSelector(BaseExporterSelector):
    # pylint: disable=too-few-public-methods

    """
    Formatter to export to Kili
    """

    @staticmethod
    def select_exporter_class(split_param: SplitOption) -> Type[BaseExporter]:
        return KiliExporter



"""
This script lists package constants.
"""


NO_ACCESS_RIGHT = (
    "[noAccessRights] It seems you do not have access to this object."
    " Please double check your credentials."
)

INPUT_TYPE = [
    "AUDIO",
    "FRAME",
    "IMAGE",
    "PDF",
    "TEXT",
    "TIME_SERIES",
    "VIDEO",
    "VIDEO_LEGACY",
]

mime_extensions = {
    "Audio": "audio/x-flac,audio/mpeg,video/mp4",
    "Csv": "text/csv",
    "Video": "video/mp4,video/x-matroska,video/3gpp,video/x-msvideo,"
    "video/x-m4v,video/quicktime,video/webm",
    "Image": "image/jpeg,image/png,image/bmp,image/gif,image/webp,image/x-icon,"
    "image/tiff,image/vnd.microsoft.icon,image/svg+xml,image/avif,image/apng",
    "Pdf": "application/pdf",
    "Text": "text/plain",
    "TimeSeries": "text/csv",
}

mime_extensions_for_IV2 = {
    "AUDIO": mime_extensions["Audio"],
    "FRAME": mime_extensions["Video"],
    "IMAGE": mime_extensions["Image"],
    "NA": "",
    "PDF": mime_extensions["Pdf"],
    "TEXT": mime_extensions["Text"],
    "TIME_SERIES": mime_extensions["Csv"],
    "URL": "",
    "VIDEO": mime_extensions["Video"],
    "VIDEO_LEGACY": mime_extensions["Csv"],
}

MUTATION_BATCH_SIZE = 100
THROTTLING_DELAY = 60 / 250



"""CLI's Project Member common functions"""

import re
import warnings
from typing import Dict, Iterable, List, Optional, cast

from kili.cli.common_args import ROLES
from kili.cli.helpers import collect_from_csv

REGEX_EMAIL = re.compile(r"([A-Za-z0-9]+[.-_])*[A-Za-z0-9]+@[A-Za-z0-9-]+(\.[A-Z|a-z]{2,})+")


def type_check_member(key, value):
    """type check value based on key"""
    if key == "email" and not re.search(REGEX_EMAIL, value):
        return f"{value} is not a valid email address, "

    if key == "role" and value not in ROLES:
        return f"{value} is not a valid role, "

    return ""


def collect_members_from_csv(csv_path: str, role: Optional[str]):
    """read a csv with to collect members and role"""
    members_to_add = collect_from_csv(
        csv_path=csv_path,
        required_columns=["email"],
        optional_columns=["role"],
        type_check_function=type_check_member,
    )

    if len(members_to_add) == 0:
        raise ValueError(f"No active member were found in csv: {csv_path}")

    if "role" in members_to_add[0].keys():
        if role is not None:
            raise ValueError(
                "--role cannot be used if the argument passed is a path to a csv file with roles"
            )
    else:
        if role is None:
            role = "LABELER"
        for member in members_to_add:
            member["role"] = role

    return members_to_add


def collect_members_from_project(kili, project_id_source: str, role: Optional[str]):
    """copy members from project of id project_id_source"""
    members = []

    if role is not None:
        raise ValueError("--role cannot be used if the argument passed is a Kili project_id")

    try:
        users = cast(
            List[Dict],
            kili.project_users(
                project_id=project_id_source,
                fields=["role", "user.email", "activated"],
                disable_tqdm=True,
            ),
        )
        for user in users:
            if user["activated"]:
                members.append({"email": user["user"]["email"], "role": user["role"]})
    except:
        # pylint: disable=raise-missing-from
        raise ValueError(f"{project_id_source} is not recognized as a Kili project_id")

    if len(members) == 0:
        raise ValueError(f"No active member were found in project with id {project_id_source}")

    return members


def collect_members_from_emails(emails: Iterable[str], role: Optional[str]):
    """collect members with email address from emails"""
    if role is None:
        role = "LABELER"
    members_to_add = []
    for email in emails:
        if re.search(REGEX_EMAIL, email):
            members_to_add.append({"email": email, "role": role})
        else:
            warnings.warn(f"{email} is not a valid email address,")

    if len(members_to_add) == 0:
        raise ValueError("No valid email adresses were provided")

    return members_to_add


def check_exclusive_options(
    csv_path: Optional[str],
    project_id_src: Optional[str],
    emails: Optional[Iterable[str]],
    all_members: Optional[bool],
) -> None:
    """Forbid mutual use of options and argument(s)"""
    if not emails:
        return None
    emails = list(emails)
    if all_members is None:
        if (csv_path is not None) + (project_id_src is not None) + (len(emails) > 0) > 1:
            raise ValueError(
                "emails arguments and options --from-csv and --from-project are exclusive."
            )
        if (csv_path is not None) + (project_id_src is not None) + (len(emails) > 0) == 0:
            raise ValueError(
                "You must either provide emails arguments or use one option from "
                "--from-csv or --from-project"
            )

    else:
        if (csv_path is not None) + (all_members is True) + (len(emails) > 0) > 1:
            raise ValueError("emails arguments and options --from-csv and --all are exclusive.")
        if (csv_path is not None) + (all_members is not None) + (len(emails) > 0) == 0:
            raise ValueError(
                "You must either provide emails arguments or use one option --from-csv or --all"
            )
    return None



"""Helpers for the asset queries"""

from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from mimetypes import guess_extension
from pathlib import Path
from typing import List, Optional

import requests
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from kili.queries.asset.exceptions import MissingPropertyError


def get_post_assets_call_process(
    download_media: bool, local_media_dir: Optional[str], project_id: str
):
    """
    Define the function to apply on assets after a paginated call of assets
    Call the download_asset_media with the right parameters if download_media is True
    Otherwise return assets without any post-processing
    """
    if download_media:
        local_dir_path = (
            Path(local_media_dir)
            if local_media_dir is not None
            else Path(f".cache/kili/projects/{project_id}/assets")
        )
        return lambda assets: download_asset_content(assets, local_dir_path)

    return lambda assets: assets


def get_file_extension_from_headers(url):
    """guess the extension of a file with the url response headers"""
    with requests.head(url, timeout=20) as header_response:
        headers = header_response.headers
        header_response.raise_for_status()
        content_type = headers["content-type"]
        return guess_extension(content_type)


def get_download_path(url: str, external_id: str, local_dir_path: Path):
    """Build the path to download a file the file in local."""
    local_path = local_dir_path / external_id
    extension = get_file_extension_from_headers(url)
    if extension is not None:
        local_path = local_path.with_suffix(extension)
    return local_path.resolve()


@retry(stop=stop_after_attempt(2), wait=wait_random(min=1, max=2))
def download_file(url: str, external_id: str, local_dir_path: Path):
    """
    Download a file by streming chunks of 1Mb
    If the file already exists in local, it does not download it
    """
    local_path = get_download_path(url, external_id, local_dir_path)
    if not local_path.is_file():
        with requests.get(url, stream=True, timeout=20) as response:
            response.raise_for_status()
            with open(local_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024 * 1024):
                    file.write(chunk)
    return local_path


def assert_required_fields_existence(asset: dict):
    """Get a list of an asset field from a generator of kili assets"""
    required_fields = ["content", "externalId"]
    for field in required_fields:
        if field not in asset.keys():
            raise MissingPropertyError(
                f"The asset does not have the {field} field. Please add it to the fields when"
                " querying assets "
            )
        content = asset["content"]
        if content == "":
            raise NotImplementedError(
                "The media of the asset is not stored in the content field. If your asset is video"
                " imported from frames or a richText asset, the media is stored in the jsonContent"
                " field. The download in local of such asset media is currently not supported."
            )


def download_asset_content(assets: List[dict], local_dir_path: Path) -> List[dict]:
    """Download assets media in local."""
    if len(assets) == 0:
        return assets
    local_dir_path.mkdir(parents=True, exist_ok=True)
    assert_required_fields_existence(assets[0])
    content_gen = (asset["content"] for asset in assets)
    external_id_gen = (asset["externalId"] for asset in assets)
    with ThreadPoolExecutor() as threads:
        path_gen = threads.map(download_file, content_gen, external_id_gen, repeat(local_dir_path))
    return [{**asset, "content": str(path)} for asset, path in zip(assets, path_gen)]



"""
Helpers for the services
"""
from pathlib import Path
from typing import Iterable, Optional, TypeVar

from kili.services.exceptions import (
    NotEnoughArgumentsSpecifiedError,
    TooManyArgumentsSpecifiedError,
)

PathLike = TypeVar("PathLike", Path, str)


def check_exclusive_options(
    csv_path: Optional[PathLike], files: Optional[Iterable[PathLike]]
) -> None:
    """Forbid mutual use of options and argument(s)"""
    if files is not None:
        files = list(files)

    if (csv_path is not None) + (files is not None and len(files) > 0) > 1:
        raise TooManyArgumentsSpecifiedError(
            "An explicit list of files and a CSV file containing label files"
            " can't be specified at the same time"
        )

    if (csv_path is not None) + (files is not None and len(files) > 0) == 0:
        raise NotEnoughArgumentsSpecifiedError(
            "You must either provide an explicit list of files or a CSV file containing a file list"
        )


def get_external_id_from_file_path(path: PathLike) -> str:
    """Return external_id from file's path
    ex: 'tree/leaf/file_name.txt- -> file_name
    """
    file_path = Path(path).parts[-1]
    return ".".join(file_path.split(".")[:-1])


def is_target_job_in_json_interface(kili, project_id: str, target_job_name: str):
    """Tell if the target job id is defined in the project's JSON interface"""
    json_interface = list(kili.projects(project_id=project_id, fields=["jsonInterface"]))[0][
        "jsonInterface"
    ]
    return target_job_name in json_interface["jobs"]



"""
Queries of asset queries
"""


def gql_api_keys(fragment):
    """
    Return the GraphQL assets query
    """
    return f"""
query($where: ApiKeyWhere!, $first: PageSize!, $skip: Int!) {{
  data: apiKeys(where: $where, skip: $skip, first: $first) {{
    {fragment}
  }}
}}
"""


GQL_API_KEYS_COUNT = """
query($where: ApiKeyWhere!) {
  data: countApiKeys(where: $where)
}
"""
